-- NEC MCM424000A32
-- this model is made of 4  4Mbyte drams
-- this model does NOT: 
-- check for proper refresh conditions, altho it recognizes refreshes

-- this version is optimized for 8 bits ie it stores 4 bytes per integer
-- using the memin and memout procedures/functions
-- this model MAY have different timing values, as it is being used to model the 4M parts
-- rlength,clength may be used to modify the row length and column length size of the
-- drams. usually use rlength smaller than clength to reduce the size of the part to
-- get it to simulate in memory

-- this is the version w/ save/load from disk capability
-- NOTE! The file names in this version require a blank as the first char
-- This is where the bank # (0-3) will be prepended in the file names
library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;

ENTITY pulse_widths_dram4p IS
    GENERIC (
        tlow: TIME := 0 ns;
        thigh: TIME := 0 ns);
    PORT (
        in0 : IN std_logic;
        out0: OUT std_logic);
END pulse_widths_dram4p;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;

ARCHITECTURE behavior OF pulse_widths_dram4p IS
BEGIN
    PROCESS (in0)
        VARIABLE prev_event : TIME := 0 ns;
    BEGIN
        IF prev_event /= 0 ns THEN
            IF in0 = '1' THEN
                ASSERT Now - prev_event > tlow
                    REPORT "Pulse width of '0' is less than " & time'image(tlow)
                    SEVERITY Warning;
            ELSIF in0 = '0' THEN
                ASSERT Now - prev_event > thigh
                    REPORT "Pulse width of '1' is less than " & time'image(thigh)
                    SEVERITY Warning;
            END IF;
        END IF;

        prev_event := Now;
    END PROCESS;

    out0 <= in0;
END behavior;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;

entity dram4p is
    generic ( rlength :integer := 2048; clength :integer := 2048; width : integer := 8;
              asize : integer :=11);
    port (
-- these activate saveing and loading mem
		save_flag,load_flag : std_logic:='0'; -- 1 for save, 1 for load
			  save_fname,load_fname : string(1 to 12);  -- bypass port file names for save, load
              rowa,cola : integer :=0;
              cnt : integer := rlength*clength;

		a : in std_logic_vector(asize-1 downto 0);
        not_ras,not_cas,not_we : in std_logic:='1';
		data : inout std_logic_vector(width-1 downto 0):="ZZZZZZZZ"
   );
end dram4p;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;
use std.textio.all;
use work.io_stuff.all;


architecture operate_dram4p of dram4p is

type mem_type is array (rlength-1 downto 0,clength-1 downto 0) of integer;
shared variable mem : mem_type;
 
FILE load_handle : TEXT;
FILE save_handle : TEXT;


    -- function evec2int (l: std_ulogic_vector)                      return natural;
    -- Translate unsigned vector into integer representation
    ---
       function evec2int (l: std_ulogic_vector)                      
             return integer is
          variable result: integer := 0;
       begin
          for t1 in l'range loop
             result := result * 2;
             if (l(t1) = '1' or l(t1) = 'H') then
                result := result + 1;
             end if;
          end loop;
          return result;
       end evec2int;


FUNCTION memout(mem : in mem_type; row,col : in integer) RETURN integer IS
variable temp : std_ulogic_vector(31 downto 0);
BEGIN
      temp := int2evec((mem(row,col/4)));
 CASE (col rem 4) IS
    WHEN 0 =>
      return evec2int(temp(7 downto 0));
    WHEN 1 =>
      return evec2int(temp(15 downto 8));
    WHEN 2 =>
      return evec2int(temp(23 downto 16));
    WHEN 3 =>
      return evec2int(temp(31 downto 24));
    WHEN OTHERS =>
 END CASE;      
END memout;

procedure memin(mem : inout mem_type; row,col : in integer; input : in integer) is
variable temp : std_ulogic_vector(31 downto 0);
begin
      temp := int2evec((mem(row,col/4)));
  
 CASE (col rem 4) IS
    WHEN 0 =>
      temp(7 downto 0) := int2evec(input)(7 downto 0);
    WHEN 1 =>
      temp(15 downto 8) := int2evec(input)(7 downto 0);
    WHEN 2 =>
      temp(23 downto 16) := int2evec(input)(7 downto 0);
    WHEN 3 =>
      temp(31 downto 24) := int2evec(input)(7 downto 0);
    WHEN OTHERS =>
 END CASE;
      mem(row,col/4) := evec2int(temp);   
  
end memin;


signal mode : std_logic;
signal cmode : std_logic;
signal last_cas_rise : TIME := 0 ns;



shared variable row_a,col_a : integer;


-- Component instantiated by Turbo autoplace on 12/08/95 at 15:49:03
COMPONENT pulse_widths_dram4p
    GENERIC (
        tlow: TIME := 0 ns;
        thigh: TIME := 0 ns);
    PORT (
        in0 : IN std_logic;
        out0: OUT std_logic);
END COMPONENT;

SIGNAL  tnot_ras,tnot_cas : Std_logic; -- temps for pulse width checks



-- values for -70 parts
constant tAA  : time := 35 ns; -- access from column address stable
constant tACP : time := 40 ns; -- access from last rising edge of CAS
constant tASR : time := 0 ns;
constant tRAH : time := 10 ns;
constant tASC : time := 0 ns;
constant tCAH : time := 15 ns;
constant tCAC : time := 20 ns;
constant tCLZ : time := 0 ns;
constant tOFZ : time := 7 ns;  -- 0 min, 15 max
constant tWCS : time := 0 ns;
constant tWCH : time := 15 ns;
constant tDHS : time := 0 ns;
constant tDHH : time := 15 ns;
constant tRAS : time := 70 ns;
constant tRP : time := 60 ns;
constant tCAS : time := 20 ns;	
constant tCP : time := 10 ns;

signal probe : time;
 

begin




-- not_ras to a setup and hold checks
assert1: for i in 0 to asize-1 generate
x : ASSERT not((not_ras='0' and not not_ras'stable) and (not a(i)'stable(tASC))) 
        REPORT "setup violation not_ras to a"
        SEVERITY warning;

y : ASSERT not((not_ras='0' and a(i)'event) and (not not_ras'stable(tCAH))) 
        REPORT "hold violation not_ras to a"
        SEVERITY warning;

end generate;

-- not_cas to a setup and hold checks
assert2: for i in 0 to asize-1 generate
x : ASSERT not((not_cas='0' and not not_cas'stable) and (not a(i)'stable(tASR))) 
        REPORT "setup violation not_cas to a"
        SEVERITY warning;

y : ASSERT not((not_cas='0' and a(i)'event) and (not not_cas'stable(tRAH))) 
        REPORT "hold violation not_cas to a"
        SEVERITY warning;

end generate;

-- not_cas to not_we setup and hold checks
g : ASSERT not((not_cas='0' and not not_cas'stable) and (not not_we'stable(tWCS))) 
        REPORT "setup violation not_cas to not_we"
        SEVERITY warning;

h : ASSERT not((not_cas='0' and not_we'event) and (not not_cas'stable(tWCH))) 
        REPORT "hold violation not_cas to not_we"
        SEVERITY warning;

-- pulse width minimums tests
pnot_ras: pulse_widths_dram4p 
    GENERIC MAP (tlow => tRAS, thigh => tRP)
    PORT MAP (in0 => not_ras, out0 => tnot_ras);
pnot_cas: pulse_widths_dram4p 
    GENERIC MAP (tlow => tCAS, thigh => tCP)
    PORT MAP (in0 => not_cas, out0 => tnot_cas);

proc_cas_rise: process
begin
    -- snap the time every cas rising edge
    wait until rising_edge(not_cas);
      last_cas_rise <= now;
end process proc_cas_rise;


proc_cas_ras: process
begin
    wait until (falling_edge(not_cas) and (not_ras='1'));
-- cas before ras operation
      cmode <= not_we;
      wait until falling_edge(not_ras) or rising_edge(not_cas);
	  if (not_ras='0') then -- this is a cas_ras op
	    CASE cmode IS
	    WHEN '1' => -- cas before ras refresh
	    WHEN OTHERS => -- 0 is undefined
          REPORT "not_we=0 - cas before ras - illegal mode"
          SEVERITY failure;
	    END CASE;
	  end if;
end process proc_cas_ras;

proc_ras_cas: process 
variable since_cas_rise : time;
begin
    wait until (falling_edge(not_ras) and (not_cas='1'));
-- ras before cas operation
-- capture mode
      mode <= not_we;
      row_a := evec2int(a);
      wait until falling_edge(not_cas) or rising_edge(not_ras);
	  if (not_cas='0') then
        col_a := evec2int(a);
        wait for 1 ps;
		  while (not_ras='0') loop
            if ((not_we='1') and (not_cas='0')) then -- read
              for i in 0 to width-1 loop
                data(i) <= TRANSPORT 'Z' after tCLZ; -- hold of old data only lasts this long
              end loop;
              since_cas_rise := now - last_cas_rise;
              if ((tACP - since_cas_rise) > tCAC) and ((tACP - since_cas_rise) > (tAA - a'last_event)) then
                -- tACP limits access
                data <= TRANSPORT int2evec(memout(mem,row_a,col_a))(width-1 downto 0) after (tACP - since_cas_rise);
			  elsif ((tAA - a'last_event) > tCAC) and ((tAA - a'last_event) > (tACP - since_cas_rise)) then 
                -- tAA limits access
                data <= TRANSPORT int2evec(memout(mem,row_a,col_a))(width-1 downto 0) after (tAA - a'last_event);
              else
                -- tCAC, the shortest of the 3, determines access
                data <= TRANSPORT int2evec(memout(mem,row_a,col_a))(width-1 downto 0) after tCAC;
              end if;
            elsif ((not_we='0') and (not_cas='0')) then -- write
              memin(mem,row_a,col_a,evec2int(data));
            end if;
            wait until (falling_edge(not_cas) or (not_ras='1') or falling_edge(not_we) or rising_edge(not_cas));
            if ((not_we='1') or (not_ras='1') or rising_edge(not_cas)) then
              for i in 0 to width-1 loop
                data(i) <= TRANSPORT 'Z' after tOFZ;
              end loop;
			end if;
			if (falling_edge(not_cas)) then
              col_a := evec2int(a);
			end if;
            wait for 1 ps;
		  end loop;
	  else -- this wud be a ras only refresh here
	  end if;
end process proc_ras_cas;

proc_byp_in: process (load_flag,rowa,cola,cnt)
variable datav,cntr : integer;
variable flag : boolean;
begin
	if (rising_edge(load_flag)) then
    cntr := 0;
    file_open(load_handle,load_fname,READ_MODE);
    for i in rowa to rlength-1 loop
      for j in cola to clength-1 loop
	    proc_vram_load(load_handle,datav,flag);
        memin(mem,i,j,datav);
        cntr := cntr+1;
        if (cnt=cntr) then exit; end if;
      end loop;
      if (cnt=cntr) then exit; end if;
    end loop;
	file_close(load_handle);
	end if;
end process proc_byp_in;

proc_byp_out: process (save_flag,rowa,cola,cnt)
variable datav,cntr : integer;
variable flag : boolean;
begin
	if (rising_edge(save_flag)) then
    cntr := 0;
    file_open(save_handle,save_fname,WRITE_MODE);
    for i in 0 to rlength-1 loop
      for j in 0 to clength-1 loop
        datav := memout(mem,i,j);
	    proc_vram_save(save_handle,datav,width/4,flag);
        cntr := cntr+1;
        if (cnt=cntr) then exit; end if;
      end loop;
      if (cnt=cntr) then exit; end if;
    end loop;
	file_close(save_handle);
	end if;
end process proc_byp_out;
	  

  
end operate_dram4p;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;

entity dram16p is
    generic ( rlength :integer := 2048; clength :integer := 2048; width : integer := 32;
              asize : integer :=11);
    port (
-- these activate saveing and loading mem
		save_flag,load_flag : std_logic:='0'; -- 1 for save, 1 for load
			  save_fname,load_fname : string(1 to 12);  -- bypass port file names for save, load
              rowa,cola : integer :=0;
              cnt : integer := rlength*clength;

		a : in std_logic_vector(asize-1 downto 0);
        not_we : in std_logic :='1';
		not_ras : in std_logic_vector(3 downto 0) :="1111";	-- note, only ras0 and ras2 are used here
		not_cas :  in std_logic_vector(3 downto 0) :="1111";
		data : inout std_logic_vector(width-1 downto 0);
		drprd : out std_logic_vector(4 downto 1):="1001"  -- defined by access speed - these are for -70
   );
end dram16p;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;
use work.io_stuff.all;


architecture operate_dram16p of dram16p is
-- Component instantiated by Turbo autoplace on 01/03/96 at 14:42:42
COMPONENT dram4p
    generic ( rlength :integer := 2048; clength :integer := 2048; width : integer := 8;
              asize : integer :=11);
    port (
-- these activate saveing and loading mem
		save_flag,load_flag : std_logic:='0'; -- 1 for save, 1 for load
			  save_fname,load_fname : string(1 to 12);  -- bypass port file names for save, load
              rowa,cola : integer :=0;
              cnt : integer := rlength*clength;

		a : in std_logic_vector(asize-1 downto 0);
        not_ras,not_cas,not_we : in std_logic:='1';
		data : inout std_logic_vector(width-1 downto 0):="ZZZZZZZZ"
   );
END COMPONENT;
type string_array is array(0 to 3) of string(1 to 12);
signal save_fnamex,load_fnamex : string_array;  -- bypass port file names for save, load
signal pass_save,pass_load : std_logic :='0';
begin

proc_ld: process
begin
	wait until rising_edge(load_flag); 
     for i in 0 to 3 loop
      load_fnamex(i) <= load_fname;
      load_fnamex(i)(1) <= v4_to_hex(i);
     end loop;
     wait for 1 ps;
      pass_load <= '1';
      wait for 1 ps;
      pass_load <= '0';
end process proc_ld;
proc_sv: process 
begin
	wait until rising_edge(save_flag);
     for i in 0 to 3 loop
      save_fnamex(i) <= save_fname;
      save_fnamex(i)(1) <= v4_to_hex(i);
     end loop;
     wait for 1 ps;
      pass_save <= '1';
      wait for 1 ps;
      pass_save <= '0';
end process proc_sv;


u0: dram4p
  generic map(rlength=>rlength,clength=>clength,width=>8,asize=>asize) 
  port map (
            save_flag => pass_save,
            load_flag => pass_load,
			save_fname => save_fnamex(0)(1 to 12),
			load_fname => load_fnamex(0)(1 to 12),
            rowa => rowa,
            cola => cola,
            cnt => cnt,
                
	a=>a,
	not_ras=>not_ras(0),
	not_cas=>not_cas(0),
	not_we=>not_we,
	data=>data(7 downto 0)
	);
u1: dram4p
  generic map(rlength=>rlength,clength=>clength,width=>8,asize=>asize) 
  port map (
            save_flag => pass_save,
            load_flag => pass_load,
			save_fname => save_fnamex(1)(1 to 12),
			load_fname => load_fnamex(1)(1 to 12),
            rowa => rowa,
            cola => cola,
            cnt => cnt,
                					
	a=>a,
	not_ras=>not_ras(0),
	not_cas=>not_cas(1),
	not_we=>not_we,
	data=>data(15 downto 8)
	);
u2: dram4p
  generic map(rlength=>rlength,clength=>clength,width=>8,asize=>asize) 
  port map (
            save_flag => pass_save,
            load_flag => pass_load,
			save_fname => save_fnamex(2)(1 to 12),
			load_fname => load_fnamex(2)(1 to 12),
            rowa => rowa,
            cola => cola,
            cnt => cnt,
                
	a=>a,
	not_ras=>not_ras(2),
	not_cas=>not_cas(2),
	not_we=>not_we,
	data=>data(23 downto 16)
	);
u3: dram4p
  generic map(rlength=>rlength,clength=>clength,width=>8,asize=>asize) 
  port map (
            save_flag => pass_save,
            load_flag => pass_load,
			save_fname => save_fnamex(3)(1 to 12),
			load_fname => load_fnamex(3)(1 to 12),
            rowa => rowa,
            cola => cola,
            cnt => cnt,
                
	a=>a,
	not_ras=>not_ras(2),
	not_cas=>not_cas(3),
	not_we=>not_we,
	data=>data(31 downto 24)
	);
end operate_dram16p;

